import struct
from enum import Enum
from typing import Dict as dict
from typing import List as list
from typing import Optional

from fit_tool.base_type import BaseType
from fit_tool.endian import Endian
from fit_tool.field_component import FieldComponent
from fit_tool.field_definition import FieldDefinition
from fit_tool.sub_field import SubField


class ArrayType(Enum):
    FIXED = 0
    VARIABLE = 1


class Field:
    encoded_values = []

    def __init__(self, field_id: int = 0, name: str = '', base_type: BaseType = BaseType.ENUM,
                 offset: float = None,
                 scale: float = None,
                 units: str = '',
                 is_accumulated: bool = False,
                 is_expanded_field=False,
                 sub_fields: list[SubField] = None,
                 components: list[FieldComponent] = None,
                 size: int = 0,
                 growable: bool = False,
                 type_name: str = '',
                 ref_field_map: dict = None,
                 array_type: ArrayType = None,
                 array_fixed_length: int = None):

        self.field_id = field_id
        self.name = name
        self.base_type = base_type
        self.offset = offset
        self.scale = scale
        self.units = units
        self.is_accumulated = is_accumulated
        self.is_expanded_field = is_expanded_field
        self.sub_fields = sub_fields if sub_fields else []
        self.components = components if components else []
        self.size = size if size else 0
        self.growable = growable
        self.type_name = type_name
        self.ref_field_map = ref_field_map
        self.array_type = array_type
        self.array_fixed_length = array_fixed_length

        self.encoded_values = [None for _ in range(Field.get_length_from_size(base_type, size))]

    @classmethod
    def from_field(cls, other):
        field = Field(field_id=other.field_id, name=other.name, base_type=other.base_type, offset=other.offset,
                      scale=other.scale, units=other.units, is_accumulated=other.is_accumulated,
                      is_expanded_field=other.is_expanded_field, sub_fields=other.sub_fields,
                      components=other.components,
                      size=other.size, growable=other.growable, type_name=other.type_name)
        field.encoded_values = other.encoded_values
        return field

    @classmethod
    def from_field_definition(cls, definition: FieldDefinition):
        field = Field(field_id=definition.field_id, name='field', base_type=definition.base_type,
                      units='', is_accumulated=False, is_expanded_field=False, size=definition.size, growable=False,
                      sub_fields=[], components=[],
                      type_name='')
        field.encoded_values = [None for _ in range(Field.get_length_from_size(definition.base_type, definition.size))]

        return field

    def get_sub_field(self, name: str = None, index: int = None) -> Optional[SubField]:
        if index is not None and 0 <= index < len(self.sub_fields):
            return self.sub_fields[index]

        if name is not None:
            for sub_field in self.sub_fields:
                if sub_field.name == name:
                    return sub_field

        return None

    def get_name(self, sub_field: SubField = None, sub_field_name: str = None, sub_field_index: int = None) -> str:
        if sub_field:
            sb = sub_field
        elif sub_field_name is not None:
            sb = self.get_sub_field(name=sub_field_name)
        elif sub_field_index is not None:
            sb = self.get_sub_field(index=sub_field_index)
        else:
            sb = None

        if sb:
            return sb.name
        else:
            return self.name

    def get_units(self, sub_field: SubField = None, sub_field_name: str = None, sub_field_index: int = None) -> str:
        if sub_field:
            sb = sub_field
        elif sub_field_name is not None:
            sb = self.get_sub_field(name=sub_field_name)
        elif sub_field_index is not None:
            sb = self.get_sub_field(index=sub_field_index)
        else:
            sb = None

        if sb:
            return sb.units
        else:
            return self.units

    def get_base_type(self, sub_field: SubField = None, sub_field_name: str = None,
                      sub_field_index: int = None) -> BaseType:
        if sub_field:
            sb = sub_field
        elif sub_field_name is not None:
            sb = self.get_sub_field(name=sub_field_name)
        elif sub_field_index is not None:
            sb = self.get_sub_field(index=sub_field_index)
        else:
            sb = None

        if sb:
            return sb.base_type
        else:
            return self.base_type

    def get_offset(self, sub_field: SubField = None, sub_field_name: str = None, sub_field_index: int = None) -> \
            Optional[float]:
        if sub_field:
            sb = sub_field
        elif sub_field_name is not None:
            sb = self.get_sub_field(name=sub_field_name)
        elif sub_field_index is not None:
            sb = self.get_sub_field(index=sub_field_index)
        else:
            sb = None

        if sb:
            return sb.offset
        else:
            return self.offset

    def get_scale(self, sub_field: SubField = None, sub_field_name: str = None, sub_field_index: int = None) -> \
            Optional[float]:
        if sub_field:
            sb = sub_field
        elif sub_field_name is not None:
            sb = self.get_sub_field(name=sub_field_name)
        elif sub_field_index is not None:
            sb = self.get_sub_field(index=sub_field_index)
        else:
            sb = None

        if sb:
            return sb.scale
        else:
            return self.scale

    def clear(self):
        self.size = 0
        self.encoded_values = []

    def is_valid(self) -> bool:
        return self.size != 0

    def is_not_valid(self) -> bool:
        return not self.is_valid()

    def get_value(self, index: int = 0, sub_field: SubField = None):
        if index < 0 or index >= len(self.encoded_values):
            return None

        encoded_value = self.encoded_values[index]
        return self.decode_value(encoded_value, sub_field)

    # Return values as a list
    def get_values(self):
        return [self.decode_value(encoded_value) for encoded_value in self.encoded_values]

    def set_values(self, values):
        for index, value in enumerate(values):
            self.set_value(index, value)

    def decode_value(self, encoded_value, sub_field: SubField = None):
        if encoded_value is None or type(encoded_value) == str:
            return encoded_value

        scale = self.get_scale(sub_field=sub_field)
        offset = self.get_offset(sub_field=sub_field)

        if (scale is None or scale == 1.0) and (offset is None or offset == 0.0):
            # no scaling
            value = encoded_value
        else:
            scale = scale if scale is not None else 1.0
            offset = offset if offset is not None else 0.0

            value = self.un_scale_offset_value(encoded_value, scale, offset)
            if self.type_name == 'date_time':
                value = round(value)

        return value

    @staticmethod
    def scale_offset_value(value: float, scale: float, offset: float) -> int:
        encoded_value = round((value + offset) * scale)
        return encoded_value

    @staticmethod
    def un_scale_offset_value(encoded_value: int, scale: float, offset: float) -> float:
        try:
            value = encoded_value / scale - offset
        except Exception as ex:
            raise ex
        return value

    @property
    def length(self) -> int:
        return len(self.encoded_values)

    def set_value(self, index: int, value, sub_field: SubField = None):
        encoded_value = self.encode_value(value, sub_field)
        self.set_encoded_value(index, encoded_value)

    def set_encoded_value(self, index: int, encoded_value, check_validity: bool = True):
        if index < 0:
            return

        if check_validity and self.base_type != BaseType.STRING:
            if not self.base_type.is_valid(encoded_value):
                raise Exception(
                    f'{self.name} encoded value {encoded_value} is not in valid range [{self.base_type.min}, {self.base_type.max}]')

        size_changed = False
        while index >= self.length:
            if (self.base_type != BaseType.STRING or self.array_type is not None) and not self.growable:
                raise Exception('Field is not growable')

            self.encoded_values.append(None)
            size_changed = True

        self.encoded_values[index] = encoded_value

        if size_changed or self.base_type == BaseType.STRING:
            new_size = self.calculate_size()
            if new_size > self.size:
                if not self.growable:
                    raise Exception('Size exceeds fixed field size of $size bytes. Consider making field growable.')
                self.size = new_size

    def encode_value(self, value, sub_field: SubField = None):
        if isinstance(value, str):
            return value

        if value is None:
            encoded_value = self.base_type.invalid_raw_value()
        else:
            scale = self.get_scale(sub_field=sub_field)
            offset = self.get_offset(sub_field=sub_field)
            if isinstance(value, Enum):
                encoded_value = value.value
            elif (scale is None or scale == 1.0) and (offset is None or offset == 0.0):
                encoded_value = int(value)
            else:
                encoded_value = self.scale_offset_value(value, scale, offset)
        return encoded_value

    def read_all_from_bytes(self, bytes_buffer: bytes, endian: Endian = Endian.LITTLE):
        if self.base_type == BaseType.STRING:
            self.read_strings_from_bytes(bytes_buffer)
        else:
            start = 0
            for index in range(len(self.encoded_values)):
                value_bytes = bytes_buffer[start:(start + self.base_type.size)]
                self.read_from_bytes(value_bytes, index, endian=endian)
                start += self.base_type.size

    def read_from_bytes(self, bytes_buffer: bytes, index: int, endian: Endian = Endian.LITTLE):
        if self.base_type == BaseType.STRING:
            raise Exception('Type cannot be string')

        encoded_value = self.get_encoded_value_from_bytes(bytes_buffer, endian=endian)
        self.set_encoded_value(index, encoded_value, check_validity=False)

    def read_strings_from_bytes(self, bytes_buffer: bytes):
        # The number of strings is dynamic and is determined by the number of null
        # terminations in the string container

        string_container = bytes_buffer.decode('utf-8')
        strings = string_container.split('\u0000')
        strings = strings[:-1]
        strings = [x for x in strings if x]
        self.encoded_values = []
        self.encoded_values.extend(strings)

    @staticmethod
    def get_length_from_size(base_type: BaseType, size: int) -> int:
        if base_type == BaseType.STRING:
            return 0 if size == 0 else 1
        else:
            length = size // base_type.size

            if length * base_type.size != size:
                raise Exception('Size is not a multiple of type: size: $size, type: $type')

            return length

    def calculate_size(self) -> int:
        if self.base_type == BaseType.STRING:
            calc_size = 0
            for value in self.encoded_values:
                if isinstance(value, str):
                    calc_size += len(value.encode('utf-8')) + 1
            return calc_size
        else:
            return self.length * self.base_type.size

    def get_encoded_value_from_bytes(self, bytes_buffer: bytes, offset: int = 0, endian: Endian = Endian.LITTLE):
        endian_symbol = '<' if endian == Endian.LITTLE else '>'

        if self.base_type in {BaseType.ENUM, BaseType.UINT8, BaseType.UINT8Z, BaseType.BYTE}:
            value, = struct.unpack_from(f'{endian_symbol}B', bytes_buffer, offset)

        elif self.base_type == BaseType.SINT8:
            value, = struct.unpack_from(f'{endian_symbol}b', bytes_buffer, offset)

        elif self.base_type == BaseType.SINT16:
            value, = struct.unpack_from(f'{endian_symbol}h', bytes_buffer, offset)

        elif self.base_type in {BaseType.UINT16, BaseType.UINT16Z}:
            value, = struct.unpack_from(f'{endian_symbol}H', bytes_buffer, offset)

        elif self.base_type == BaseType.SINT32:
            value, = struct.unpack_from(f'{endian_symbol}i', bytes_buffer, offset)

        elif self.base_type in {BaseType.UINT32, BaseType.UINT32Z}:
            value, = struct.unpack_from(f'{endian_symbol}I', bytes_buffer, offset)

        elif self.base_type == BaseType.SINT64:
            value, = struct.unpack_from(f'{endian_symbol}q', bytes_buffer, offset)

        elif self.base_type in {BaseType.UINT64, BaseType.UINT64Z}:
            value, = struct.unpack_from(f'{endian_symbol}Q', bytes_buffer, offset)

        elif self.base_type == BaseType.FLOAT32:
            value, = struct.unpack_from(f'{endian_symbol}f', bytes_buffer, offset)

        elif self.base_type == BaseType.FLOAT64:
            value, = struct.unpack_from(f'{endian_symbol}d', bytes_buffer, offset)

        elif self.base_type == BaseType.STRING:
            length = len(bytes_buffer) - 1 - offset
            value, = struct.unpack_from(f'{length}s', bytes_buffer, offset)
            value = value.decode('utf-8')
        else:
            value = None

        return value

    def encoded_value_to_bytes(self, encoded_value, endian: Endian = Endian.LITTLE) -> bytes:
        if encoded_value is None:
            raise Exception('Value cannot be None')

        if self.base_type == BaseType.STRING:
            return encoded_value.encode('utf-8') + b'\0'

        endian_symbol = '<' if endian == Endian.LITTLE else '>'
        bytes_buffer = bytearray(b'\0' * self.base_type.size)
        if self.base_type in {BaseType.ENUM, BaseType.UINT8, BaseType.UINT8Z, BaseType.BYTE}:
            struct.pack_into('B', bytes_buffer, 0, encoded_value)

        elif self.base_type == BaseType.SINT8:
            struct.pack_into('b', bytes_buffer, 0, encoded_value)

        elif self.base_type == BaseType.SINT16:
            struct.pack_into(f'{endian_symbol}h', bytes_buffer, 0, encoded_value)

        elif self.base_type in {BaseType.UINT16, BaseType.UINT16Z}:
            struct.pack_into(f'{endian_symbol}H', bytes_buffer, 0, encoded_value)

        elif self.base_type == BaseType.SINT32:
            struct.pack_into(f'{endian_symbol}i', bytes_buffer, 0, encoded_value)

        elif self.base_type in {BaseType.UINT32, BaseType.UINT32Z}:
            struct.pack_into(f'{endian_symbol}I', bytes_buffer, 0, encoded_value)

        elif self.base_type == BaseType.SINT64:
            struct.pack_into(f'{endian_symbol}q', bytes_buffer, 0, encoded_value)

        elif self.base_type in {BaseType.UINT64, BaseType.UINT64Z}:
            struct.pack_into(f'{endian_symbol}Q', bytes_buffer, 0, encoded_value)

        elif self.base_type == BaseType.FLOAT32:
            struct.pack_into(f'{endian_symbol}f', bytes_buffer, 0, encoded_value)

        elif self.base_type == BaseType.FLOAT64:
            struct.pack_into(f'{endian_symbol}d', bytes_buffer, 0, encoded_value)

        return bytes_buffer

    def to_bytes(self, endian: Endian = Endian.LITTLE) -> bytes:
        bytes_buffer = b''
        for value in self.encoded_values:
            bytes_buffer += self.encoded_value_to_bytes(value, endian=endian)

        # sometimes subfields or strings can be less than the allocated field size,
        # so we pad the buffer with 0's to meet the size requirement.
        bytes_buffer = bytes_buffer.ljust(self.size, b'\0')

        return bytes_buffer

    def get_valid_sub_field(self, fields: list) -> Optional[SubField]:
        if not self.sub_fields:
            return None

        for sub_field in self.sub_fields:
            if sub_field.is_valid(fields):
                return sub_field

        return None

    def to_row(self, sub_field: SubField = None) -> list:
        row = []
        values = []

        for i in range(len(self.encoded_values)):
            values.append(self.get_value(index=i, sub_field=sub_field))

        row.append(sub_field.name if sub_field else self.name)

        if len(values) == 1:
            row.append(values[0])
        else:
            values_str = '[' + ','.join([str(v) for v in values]) + ']'
            row.append(values_str)

        row.append(sub_field.units if sub_field else self.units)

        return row



# Autogenerated. Do not modify.
#
# Profile: 21.60
from typing import List as list
from typing import Optional

from fit_tool.base_type import BaseType
from fit_tool.data_message import DataMessage
from fit_tool.definition_message import DefinitionMessage
from fit_tool.developer_field import DeveloperField
from fit_tool.endian import Endian
from fit_tool.field import Field


class WeightScaleMessage(DataMessage):
    ID = 30
    NAME = 'weight_scale'

    @staticmethod
    def __get_field_size(definition_message: DefinitionMessage, field_id: int) -> int:
        size = 0
        if definition_message:
            field_definition = definition_message.get_field_definition(field_id)
            if field_definition:
                size = field_definition.size

        return size

    def __init__(self, definition_message=None, developer_fields=None, local_id: int = 0,
                 endian: Endian = Endian.LITTLE):
        super().__init__(name=WeightScaleMessage.NAME,
                         global_id=WeightScaleMessage.ID,
                         local_id=definition_message.local_id if definition_message else local_id,
                         endian=definition_message.endian if definition_message else endian,
                         definition_message=definition_message,
                         developer_fields=developer_fields,
                         fields=[
                             TimestampField(
                                 size=self.__get_field_size(definition_message, TimestampField.ID),
                                 growable=definition_message is None),
                             WeightScaleWeightField(
                                 size=self.__get_field_size(definition_message, WeightScaleWeightField.ID),
                                 growable=definition_message is None),
                             WeightScalePercentFatField(
                                 size=self.__get_field_size(definition_message, WeightScalePercentFatField.ID),
                                 growable=definition_message is None),
                             WeightScalePercentHydrationField(
                                 size=self.__get_field_size(definition_message, WeightScalePercentHydrationField.ID),
                                 growable=definition_message is None),
                             WeightScaleVisceralFatMassField(
                                 size=self.__get_field_size(definition_message, WeightScaleVisceralFatMassField.ID),
                                 growable=definition_message is None),
                             WeightScaleBoneMassField(
                                 size=self.__get_field_size(definition_message, WeightScaleBoneMassField.ID),
                                 growable=definition_message is None),
                             WeightScaleMuscleMassField(
                                 size=self.__get_field_size(definition_message, WeightScaleMuscleMassField.ID),
                                 growable=definition_message is None),
                             WeightScaleBasalMetField(
                                 size=self.__get_field_size(definition_message, WeightScaleBasalMetField.ID),
                                 growable=definition_message is None),
                             WeightScalePhysiqueRatingField(
                                 size=self.__get_field_size(definition_message, WeightScalePhysiqueRatingField.ID),
                                 growable=definition_message is None),
                             WeightScaleActiveMetField(
                                 size=self.__get_field_size(definition_message, WeightScaleActiveMetField.ID),
                                 growable=definition_message is None),
                             WeightScaleMetabolicAgeField(
                                 size=self.__get_field_size(definition_message, WeightScaleMetabolicAgeField.ID),
                                 growable=definition_message is None),
                             WeightScaleVisceralFatRatingField(
                                 size=self.__get_field_size(definition_message, WeightScaleVisceralFatRatingField.ID),
                                 growable=definition_message is None),
                             WeightScaleUserProfileIndexField(
                                 size=self.__get_field_size(definition_message, WeightScaleUserProfileIndexField.ID),
                                 growable=definition_message is None)
                         ])

        self.growable = self.definition_message is None

    @classmethod
    def from_bytes(cls, definition_message: DefinitionMessage, developer_fields: list[DeveloperField],
                   bytes_buffer: bytes, offset: int = 0):
        message = cls(definition_message=definition_message, developer_fields=developer_fields)
        message.read_from_bytes(bytes_buffer, offset)
        return message

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @property
    def timestamp(self) -> Optional[int]:
        field = self.get_field(TimestampField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @timestamp.setter
    def timestamp(self, value: int):
        field = self.get_field(TimestampField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def weight(self) -> Optional[float]:
        field = self.get_field(WeightScaleWeightField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @weight.setter
    def weight(self, value: float):
        field = self.get_field(WeightScaleWeightField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def percent_fat(self) -> Optional[float]:
        field = self.get_field(WeightScalePercentFatField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @percent_fat.setter
    def percent_fat(self, value: float):
        field = self.get_field(WeightScalePercentFatField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def percent_hydration(self) -> Optional[float]:
        field = self.get_field(WeightScalePercentHydrationField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @percent_hydration.setter
    def percent_hydration(self, value: float):
        field = self.get_field(WeightScalePercentHydrationField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def visceral_fat_mass(self) -> Optional[float]:
        field = self.get_field(WeightScaleVisceralFatMassField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @visceral_fat_mass.setter
    def visceral_fat_mass(self, value: float):
        field = self.get_field(WeightScaleVisceralFatMassField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bone_mass(self) -> Optional[float]:
        field = self.get_field(WeightScaleBoneMassField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bone_mass.setter
    def bone_mass(self, value: float):
        field = self.get_field(WeightScaleBoneMassField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def muscle_mass(self) -> Optional[float]:
        field = self.get_field(WeightScaleMuscleMassField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @muscle_mass.setter
    def muscle_mass(self, value: float):
        field = self.get_field(WeightScaleMuscleMassField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def basal_met(self) -> Optional[float]:
        field = self.get_field(WeightScaleBasalMetField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @basal_met.setter
    def basal_met(self, value: float):
        field = self.get_field(WeightScaleBasalMetField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def physique_rating(self) -> Optional[int]:
        field = self.get_field(WeightScalePhysiqueRatingField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @physique_rating.setter
    def physique_rating(self, value: int):
        field = self.get_field(WeightScalePhysiqueRatingField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def active_met(self) -> Optional[float]:
        field = self.get_field(WeightScaleActiveMetField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @active_met.setter
    def active_met(self, value: float):
        field = self.get_field(WeightScaleActiveMetField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def metabolic_age(self) -> Optional[int]:
        field = self.get_field(WeightScaleMetabolicAgeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @metabolic_age.setter
    def metabolic_age(self, value: int):
        field = self.get_field(WeightScaleMetabolicAgeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def visceral_fat_rating(self) -> Optional[int]:
        field = self.get_field(WeightScaleVisceralFatRatingField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @visceral_fat_rating.setter
    def visceral_fat_rating(self, value: int):
        field = self.get_field(WeightScaleVisceralFatRatingField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def user_profile_index(self) -> Optional[int]:
        field = self.get_field(WeightScaleUserProfileIndexField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @user_profile_index.setter
    def user_profile_index(self, value: int):
        field = self.get_field(WeightScaleUserProfileIndexField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)


class TimestampField(Field):
    ID = 253

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='timestamp',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=-631065600000,
            scale=0.001,
            size=size,
            units='ms',
            type_name='date_time',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScaleWeightField(Field):
    ID = 0

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='weight',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='kg',
            type_name='weight',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScalePercentFatField(Field):
    ID = 1

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='percent_fat',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='%',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScalePercentHydrationField(Field):
    ID = 2

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='percent_hydration',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='%',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScaleVisceralFatMassField(Field):
    ID = 3

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='visceral_fat_mass',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='kg',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScaleBoneMassField(Field):
    ID = 4

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bone_mass',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='kg',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScaleMuscleMassField(Field):
    ID = 5

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='muscle_mass',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='kg',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScaleBasalMetField(Field):
    ID = 7

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='basal_met',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=4,
            size=size,
            units='kcal/day',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScalePhysiqueRatingField(Field):
    ID = 8

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='physique_rating',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScaleActiveMetField(Field):
    ID = 9

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='active_met',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=4,
            size=size,
            units='kcal/day',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScaleMetabolicAgeField(Field):
    ID = 10

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='metabolic_age',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            units='years',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScaleVisceralFatRatingField(Field):
    ID = 11

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='visceral_fat_rating',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class WeightScaleUserProfileIndexField(Field):
    ID = 12

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='user_profile_index',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )



import struct
from typing import Dict as dict
from typing import List as list
from typing import Optional

from fit_tool.developer_field import DeveloperField
from fit_tool.developer_field_definition import DeveloperFieldDefinition
from fit_tool.endian import Endian
from fit_tool.field_definition import FieldDefinition
from fit_tool.message import Message
from fit_tool.utils.logging import logger


class DefinitionMessage(Message):

    def __init__(self, local_id: int = 0, global_id: int = 0, endian: Endian = Endian.LITTLE,
                 field_definitions: list[FieldDefinition] = None,
                 developer_field_definitions: list[DeveloperFieldDefinition] = None
                 ):
        super().__init__(local_id=local_id, global_id=global_id,
                         size=DefinitionMessage.calculate_size(field_definitions, developer_field_definitions),
                         endian=endian)

        self.field_definitions = field_definitions if field_definitions else []
        self.developer_field_definitions = developer_field_definitions if developer_field_definitions else []

    @property
    def defined_data_size(self) -> int:
        size = 0
        for field_definition in self.field_definitions:
            size += field_definition.size

        for developer_field_definition in self.developer_field_definitions:
            size += developer_field_definition.size

        return size

    def has_developer_fields(self) -> bool:
        return len(self.developer_field_definitions) > 0

    def get_field_definition(self, field_id: int) -> Optional[FieldDefinition]:
        return next((x for x in self.field_definitions if x.field_id == field_id), None)

    def remove_field(self, field_id: int):
        field_definition = self.get_field_definition(field_id)
        if field_definition:
            self.field_definitions.remove(field_definition)
            self.size = DefinitionMessage.calculate_size(self.field_definitions, self.developer_field_definitions)

    def remove_developer_field(self, developer_data_index: int, field_id: int):
        field_definition = self.get_developer_field_definition(developer_data_index, field_id)
        if field_definition:
            self.developer_field_definitions.remove(field_definition)
            self.size = DefinitionMessage.calculate_size(self.field_definitions, self.developer_field_definitions)

    def add_field_definition(self, definition: FieldDefinition):
        self.field_definitions.append(definition)

    def get_developer_field_definition(self, developer_data_index: int, field_id: int) \
            -> Optional[DeveloperFieldDefinition]:
        return next((x for x in self.developer_field_definitions if
                     x.developer_data_index == developer_data_index and x.field_id == field_id), None)

    def add_developer_field_definition(self, definition: DeveloperFieldDefinition):
        self.developer_field_definitions.append(definition)

    def to_row(self) -> list:
        from fit_tool.profile.messages.message_factory import MessageFactory
        values = []
        message = MessageFactory.from_definition(self, [])
        values.append(message.name)

        for field_definition in self.field_definitions:
            field = message.get_field(field_definition.field_id)

            if not field:
                logger.warning(f'Field id:{field_definition.field_id} could not be found in message: {message.name}')
                continue

            values.append(field.name)
            values.append(field_definition.size)
            values.append('bytes')

        for field_definition in self.developer_field_definitions:
            field = self.get_developer_field_definition(field_definition.developer_data_index,
                                                        field_definition.field_id)

            if not field:
                logger.warning(f'Field id:{field_definition.field_id} could not be found in message: {message.name}')
                continue

            values.append(f'[{field.developer_data_index},{field.field_id}]');
            values.append(field.size)
            ''
            values.append('bytes')

        return values

    def to_bytes(self) -> bytes:
        bytes_buffer = bytearray()
        endian_symbol = '<' if self.endian == Endian.LITTLE else '>'

        # reserved
        bytes_buffer.append(0)

        # architecture
        bytes_buffer.append(0 if self.endian == Endian.LITTLE else 1)

        # global id
        buffer = struct.pack(f'{endian_symbol}H', self.global_id)
        bytes_buffer += buffer

        # field count
        bytes_buffer.append(len(self.field_definitions))

        # field definitions
        for fd in self.field_definitions:
            bytes_buffer += fd.to_bytes()

        # developer field definitions
        if self.developer_field_definitions:
            bytes_buffer.append(len(self.developer_field_definitions))

            # developer field definitions
            for fd in self.developer_field_definitions:
                bytes_buffer += fd.to_bytes()

        return bytes(bytes_buffer)

    def get_developer_fields(self, developer_fields_by_data_index: dict) -> list[DeveloperField]:
        developer_fields = []

        for field_definition in self.developer_field_definitions:
            developer_field = developer_fields_by_data_index[field_definition.developer_data_index][
                field_definition.field_id]
            if developer_field:
                sized_developer_field = DeveloperField.from_developer_field(developer_field, size=field_definition.size)
                developer_fields.append(sized_developer_field)

        return developer_fields

    @classmethod
    def from_bytes(cls, bytes_buffer: bytes, offset: int = 0, has_developer_fields: bool = False):

        # reserved
        offset += 1

        # architecture
        value, = struct.unpack_from('B', bytes_buffer, offset)
        endian = Endian(value)
        endian_symbol = '<' if endian == Endian.LITTLE else '>'
        offset += 1

        # global id
        global_id, = struct.unpack_from(f'{endian_symbol}H', bytes_buffer, offset)
        offset += 2

        # number of fields
        field_count, = struct.unpack_from('B', bytes_buffer, offset)
        offset += 1

        # Field definitions;
        field_definitions = []
        field_definition_size = FieldDefinition.field_definition_size()
        for i in range(field_count):
            fd_bytes = bytes_buffer[offset:offset + field_definition_size]
            field_definition = FieldDefinition.from_bytes(fd_bytes)
            field_definitions.append(field_definition)
            offset += field_definition_size

        # Developer Field definitions;
        developer_field_definitions = []
        if has_developer_fields:
            dev_field_count, = struct.unpack_from('B', bytes_buffer, offset)
            offset += 1

            developer_field_definition_size = DeveloperFieldDefinition.field_definition_size()
            for i in range(dev_field_count):
                fd_bytes = bytes_buffer[offset:offset + developer_field_definition_size]
                field_definition = DeveloperFieldDefinition.from_bytes(fd_bytes)
                developer_field_definitions.append(field_definition)
                offset += developer_field_definition_size

        return cls(endian=endian, global_id=global_id, field_definitions=field_definitions,
                   developer_field_definitions=developer_field_definitions)

    @classmethod
    def from_data_message(cls, data_message, min_string_size: int = 0):

        if data_message.definition_message:
            return data_message.definition_message

        field_definitions = []

        for field in data_message.fields:
            if field.is_valid():
                field_definitions.append(FieldDefinition.from_field(field, min_string_size=min_string_size))

        developer_field_definitions = []
        for field in data_message.developer_fields:
            if field.is_valid():
                developer_field_definitions.append(
                    DeveloperFieldDefinition.from_field(field, min_string_size=min_string_size))

        return cls(
            endian=data_message.endian,
            global_id=data_message.global_id,
            local_id=data_message.local_id,
            field_definitions=field_definitions,
            developer_field_definitions=developer_field_definitions)

    @staticmethod
    def calculate_size(field_definitions: list[FieldDefinition],
                       developer_field_definitions: list[DeveloperFieldDefinition]):
        field_definitions_count = len(field_definitions) if field_definitions else 0
        developer_field_definitions_count = len(developer_field_definitions) if developer_field_definitions else 0

        if developer_field_definitions_count == 0:
            return 5 + FieldDefinition.field_definition_size() * field_definitions_count + DeveloperFieldDefinition.field_definition_size() * developer_field_definitions_count
        else:
            return 5 + FieldDefinition.field_definition_size() * field_definitions_count + 1 + DeveloperFieldDefinition.field_definition_size() * developer_field_definitions_count

    def supports(self, other) -> bool:
        if self.global_id != other.global_id:
            return False

        if self.local_id != other.local_id:
            return False

        if self.endian != other.endian:
            return False

        if len(self.field_definitions) != len(other.field_definitions):
            return False

        for i in range(len(self.field_definitions)):
            field_definition = self.field_definitions[i]
            other_field_definition = other.field_definitions[i]

            if field_definition.field_id != other_field_definition.field_id:
                return False

            if field_definition.base_type != other_field_definition.base_type:
                return False

            if field_definition.size < other_field_definition.size:
                return False

        if len(self.developer_field_definitions) != len(other.developer_field_definitions):
            return False

        for i in range(len(self.developer_field_definitions)):
            field_definition = self.developer_field_definitions[i]
            other_field_definition = other.developer_field_definitions[i]

            if field_definition.field_id != other_field_definition.field_id:
                return False

            if field_definition.size < other_field_definition.size:
                return False

        return True



from typing import Dict as dict
from typing import List as list

from fit_tool.data_message import DataMessage
from fit_tool.definition_message import DefinitionMessage
from fit_tool.developer_field import DeveloperField
from fit_tool.message import Message


class RecordHeader:
    """The record header indicates whether the record content contains a definition message, a normal data message or a
     compressed timestamp data message. The record header also has a Local Message Type field that references the local
     message in the data record to its global FIT message."""

    HEADER_SIZE = 1
    IS_TIME_COMPRESSED_BIT_MASK = 0x80  # bits 7
    IS_DEFINITION_BIT_MASK = 0x40  # bits 6
    HAS_DEVELOPER_FIELDS_BIT_MASK = 0x20  # bits 5

    NORMAL_LOCAL_ID_BIT_MASK = 0x0f  # bits 0-3
    TIME_COMPRESSED_LOCAL_ID_BIT_MASK = 0x60  # bits 5-6 (0b0110 0000)
    TIME_OFFSET_BIT_MASK = 0x1f  # bits 0-4 (0b0001 1111)

    MAX_NORMAL_LOCAL_ID = 15
    MAX_TIME_COMPRESSED_LOCAL_ID = 3

    def __init__(self, is_time_compressed: bool = False, is_definition: bool = True, has_developer_fields: bool = False,
                 local_id: int = 0, time_offset_seconds: int = 0):
        # time compressed if true, otherwise normal header
        self.is_time_compressed = is_time_compressed

        # if true body contains a definition message, otherwise  body contains a data message
        self.is_definition = is_definition

        # if true messages also contain developer fields
        self.has_developer_fields = has_developer_fields
        self.local_id = local_id

        # only valid if this is a time compressed header
        self.time_offset_seconds = time_offset_seconds

    @property
    def size(self):
        return RecordHeader.HEADER_SIZE

    @classmethod
    def from_message(cls, message: Message):

        if isinstance(message, DefinitionMessage):
            has_developer_fields = message.has_developer_fields()
            return cls(
                has_developer_fields=has_developer_fields,
                local_id=message.local_id)
        else:
            return cls(is_definition=False, local_id=message.local_id)

    @classmethod
    def from_bytes(cls, bytes_buffer: bytes, offset: int = 0):
        byte = bytes_buffer[offset]

        is_time_compressed = (byte & cls.IS_TIME_COMPRESSED_BIT_MASK) == cls.IS_TIME_COMPRESSED_BIT_MASK
        if is_time_compressed:
            local_id = (byte & cls.TIME_COMPRESSED_LOCAL_ID_BIT_MASK) >> 5
            time_offset_seconds = byte & cls.TIME_OFFSET_BIT_MASK

            return cls(is_time_compressed=True, local_id=local_id, time_offset_seconds=time_offset_seconds)
        else:
            is_definition = (byte & cls.IS_DEFINITION_BIT_MASK) == cls.IS_DEFINITION_BIT_MASK
            has_developer_fields = (byte & cls.HAS_DEVELOPER_FIELDS_BIT_MASK) == cls.HAS_DEVELOPER_FIELDS_BIT_MASK
            local_id = byte & cls.NORMAL_LOCAL_ID_BIT_MASK

            return cls(is_definition=is_definition, has_developer_fields=has_developer_fields, local_id=local_id)

    def to_bytes(self):

        byte = 0x00

        if self.is_time_compressed:
            byte |= RecordHeader.IS_TIME_COMPRESSED_BIT_MASK
            byte |= (self.local_id << 5) & RecordHeader.TIME_COMPRESSED_LOCAL_ID_BIT_MASK
            byte |= self.time_offset_seconds & RecordHeader.TIME_OFFSET_BIT_MASK
        else:
            # normal header
            if self.is_definition:
                byte |= RecordHeader.IS_DEFINITION_BIT_MASK
                if self.has_developer_fields:
                    byte |= RecordHeader.HAS_DEVELOPER_FIELDS_BIT_MASK

            byte |= (self.local_id & RecordHeader.NORMAL_LOCAL_ID_BIT_MASK)

        return byte.to_bytes(1, 'little')

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def to_row(self) -> []:
        header_string = 'Definition' if self.is_definition else 'Data'
        return [header_string, self.local_id]


class Record:

    def __init__(self, header: RecordHeader, message: Message):
        self.header = header
        self.message = message

    @property
    def local_id(self) -> int:
        return self.header.local_id

    @property
    def is_definition(self) -> bool:
        return self.header.is_definition

    @property
    def size(self) -> int:
        return self.header.size + self.message.size

    @classmethod
    def from_message(cls, message: Message):
        header = RecordHeader.from_message(message)
        return cls(header, message)

    @classmethod
    def from_bytes(cls, definition_messages: dict[int, DefinitionMessage], bytes_buffer: bytes, offset: int = 0,
                   developer_fields_by_data_index: dict[int, dict[int, DeveloperField]] = None):
        header = RecordHeader.from_bytes(bytes_buffer, offset=offset)
        offset += header.size

        if header.is_definition:
            message = DefinitionMessage.from_bytes(bytes_buffer, offset=offset,
                                                   has_developer_fields=header.has_developer_fields)
        else:
            definition_message = definition_messages[header.local_id]

            if not definition_message:
                raise Exception(f'DefinitionMessage not defined for local_id: {header.local_id}')

            if developer_fields_by_data_index:
                developer_fields = definition_message.get_developer_fields(developer_fields_by_data_index)
            else:
                developer_fields = []
            message = DataMessage.from_bytes(definition_message, developer_fields, bytes_buffer, offset=offset)

        return cls(header, message)

    def to_bytes(self):
        return self.header.to_bytes() + self.message.to_bytes()

    def to_row(self) -> list:
        row = []
        row.extend(self.header.to_row())
        row.extend(self.message.to_row())
        return row

    def defined_size(self, definition_message: DefinitionMessage = None) -> int:
        if self.header.is_definition:
            return self.size
        elif definition_message:
            return self.header.size + definition_message.defined_data_size
        else:
            return 0



class FieldComponent:
    def __init__(self, field_id: int, accumulate: bool, bits: int, scale: float, offset: float):
        self.field_id = field_id
        self.accumulate = accumulate
        self.bits = bits
        self.scale = scale
        self.offset = offset



# Autogenerated. Do not modify.
#
# Profile: 21.60
from typing import List as list
from typing import Optional

from fit_tool.base_type import BaseType
from fit_tool.data_message import DataMessage
from fit_tool.definition_message import DefinitionMessage
from fit_tool.developer_field import DeveloperField
from fit_tool.endian import Endian
from fit_tool.field import Field
from fit_tool.profile.profile_type import *


class BikeProfileMessage(DataMessage):
    ID = 6
    NAME = 'bike_profile'

    @staticmethod
    def __get_field_size(definition_message: DefinitionMessage, field_id: int) -> int:
        size = 0
        if definition_message:
            field_definition = definition_message.get_field_definition(field_id)
            if field_definition:
                size = field_definition.size

        return size

    def __init__(self, definition_message=None, developer_fields=None, local_id: int = 0,
                 endian: Endian = Endian.LITTLE):
        super().__init__(name=BikeProfileMessage.NAME,
                         global_id=BikeProfileMessage.ID,
                         local_id=definition_message.local_id if definition_message else local_id,
                         endian=definition_message.endian if definition_message else endian,
                         definition_message=definition_message,
                         developer_fields=developer_fields,
                         fields=[
                             MessageIndexField(
                                 size=self.__get_field_size(definition_message, MessageIndexField.ID),
                                 growable=definition_message is None),
                             BikeProfileNameField(
                                 size=self.__get_field_size(definition_message, BikeProfileNameField.ID),
                                 growable=definition_message is None),
                             BikeProfileSportField(
                                 size=self.__get_field_size(definition_message, BikeProfileSportField.ID),
                                 growable=definition_message is None),
                             BikeProfileSubSportField(
                                 size=self.__get_field_size(definition_message, BikeProfileSubSportField.ID),
                                 growable=definition_message is None),
                             BikeProfileOdometerField(
                                 size=self.__get_field_size(definition_message, BikeProfileOdometerField.ID),
                                 growable=definition_message is None),
                             BikeProfileBikeSpdAntIdField(
                                 size=self.__get_field_size(definition_message, BikeProfileBikeSpdAntIdField.ID),
                                 growable=definition_message is None),
                             BikeProfileBikeCadAntIdField(
                                 size=self.__get_field_size(definition_message, BikeProfileBikeCadAntIdField.ID),
                                 growable=definition_message is None),
                             BikeProfileBikeSpdcadAntIdField(
                                 size=self.__get_field_size(definition_message, BikeProfileBikeSpdcadAntIdField.ID),
                                 growable=definition_message is None),
                             BikeProfileBikePowerAntIdField(
                                 size=self.__get_field_size(definition_message, BikeProfileBikePowerAntIdField.ID),
                                 growable=definition_message is None),
                             BikeProfileCustomWheelsizeField(
                                 size=self.__get_field_size(definition_message, BikeProfileCustomWheelsizeField.ID),
                                 growable=definition_message is None),
                             BikeProfileAutoWheelsizeField(
                                 size=self.__get_field_size(definition_message, BikeProfileAutoWheelsizeField.ID),
                                 growable=definition_message is None),
                             BikeProfileBikeWeightField(
                                 size=self.__get_field_size(definition_message, BikeProfileBikeWeightField.ID),
                                 growable=definition_message is None),
                             BikeProfilePowerCalFactorField(
                                 size=self.__get_field_size(definition_message, BikeProfilePowerCalFactorField.ID),
                                 growable=definition_message is None),
                             BikeProfileAutoWheelCalField(
                                 size=self.__get_field_size(definition_message, BikeProfileAutoWheelCalField.ID),
                                 growable=definition_message is None),
                             BikeProfileAutoPowerZeroField(
                                 size=self.__get_field_size(definition_message, BikeProfileAutoPowerZeroField.ID),
                                 growable=definition_message is None),
                             BikeProfileIdField(
                                 size=self.__get_field_size(definition_message, BikeProfileIdField.ID),
                                 growable=definition_message is None),
                             BikeProfileSpdEnabledField(
                                 size=self.__get_field_size(definition_message, BikeProfileSpdEnabledField.ID),
                                 growable=definition_message is None),
                             BikeProfileCadEnabledField(
                                 size=self.__get_field_size(definition_message, BikeProfileCadEnabledField.ID),
                                 growable=definition_message is None),
                             BikeProfileSpdcadEnabledField(
                                 size=self.__get_field_size(definition_message, BikeProfileSpdcadEnabledField.ID),
                                 growable=definition_message is None),
                             BikeProfilePowerEnabledField(
                                 size=self.__get_field_size(definition_message, BikeProfilePowerEnabledField.ID),
                                 growable=definition_message is None),
                             BikeProfileCrankLengthField(
                                 size=self.__get_field_size(definition_message, BikeProfileCrankLengthField.ID),
                                 growable=definition_message is None),
                             BikeProfileEnabledField(
                                 size=self.__get_field_size(definition_message, BikeProfileEnabledField.ID),
                                 growable=definition_message is None),
                             BikeProfileBikeSpdAntIdTransTypeField(
                                 size=self.__get_field_size(definition_message,
                                                            BikeProfileBikeSpdAntIdTransTypeField.ID),
                                 growable=definition_message is None),
                             BikeProfileBikeCadAntIdTransTypeField(
                                 size=self.__get_field_size(definition_message,
                                                            BikeProfileBikeCadAntIdTransTypeField.ID),
                                 growable=definition_message is None),
                             BikeProfileBikeSpdcadAntIdTransTypeField(
                                 size=self.__get_field_size(definition_message,
                                                            BikeProfileBikeSpdcadAntIdTransTypeField.ID),
                                 growable=definition_message is None),
                             BikeProfileBikePowerAntIdTransTypeField(
                                 size=self.__get_field_size(definition_message,
                                                            BikeProfileBikePowerAntIdTransTypeField.ID),
                                 growable=definition_message is None),
                             BikeProfileOdometerRolloverField(
                                 size=self.__get_field_size(definition_message, BikeProfileOdometerRolloverField.ID),
                                 growable=definition_message is None),
                             BikeProfileFrontGearNumField(
                                 size=self.__get_field_size(definition_message, BikeProfileFrontGearNumField.ID),
                                 growable=definition_message is None),
                             BikeProfileFrontGearField(
                                 size=self.__get_field_size(definition_message, BikeProfileFrontGearField.ID),
                                 growable=definition_message is None),
                             BikeProfileRearGearNumField(
                                 size=self.__get_field_size(definition_message, BikeProfileRearGearNumField.ID),
                                 growable=definition_message is None),
                             BikeProfileRearGearField(
                                 size=self.__get_field_size(definition_message, BikeProfileRearGearField.ID),
                                 growable=definition_message is None),
                             BikeProfileShimanoDi2EnabledField(
                                 size=self.__get_field_size(definition_message, BikeProfileShimanoDi2EnabledField.ID),
                                 growable=definition_message is None)
                         ])

        self.growable = self.definition_message is None

    @classmethod
    def from_bytes(cls, definition_message: DefinitionMessage, developer_fields: list[DeveloperField],
                   bytes_buffer: bytes, offset: int = 0):
        message = cls(definition_message=definition_message, developer_fields=developer_fields)
        message.read_from_bytes(bytes_buffer, offset)
        return message

    @property
    def message_index(self) -> Optional[int]:
        field = self.get_field(MessageIndexField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @message_index.setter
    def message_index(self, value: int):
        field = self.get_field(MessageIndexField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_profile_name(self) -> Optional[str]:
        field = self.get_field(BikeProfileNameField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_profile_name.setter
    def bike_profile_name(self, value: str):
        field = self.get_field(BikeProfileNameField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def sport(self) -> Optional[Sport]:
        field = self.get_field(BikeProfileSportField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @sport.setter
    def sport(self, value: Sport):
        field = self.get_field(BikeProfileSportField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def sub_sport(self) -> Optional[SubSport]:
        field = self.get_field(BikeProfileSubSportField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @sub_sport.setter
    def sub_sport(self, value: SubSport):
        field = self.get_field(BikeProfileSubSportField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def odometer(self) -> Optional[float]:
        field = self.get_field(BikeProfileOdometerField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @odometer.setter
    def odometer(self, value: float):
        field = self.get_field(BikeProfileOdometerField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_spd_ant_id(self) -> Optional[int]:
        field = self.get_field(BikeProfileBikeSpdAntIdField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_spd_ant_id.setter
    def bike_spd_ant_id(self, value: int):
        field = self.get_field(BikeProfileBikeSpdAntIdField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_cad_ant_id(self) -> Optional[int]:
        field = self.get_field(BikeProfileBikeCadAntIdField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_cad_ant_id.setter
    def bike_cad_ant_id(self, value: int):
        field = self.get_field(BikeProfileBikeCadAntIdField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_spdcad_ant_id(self) -> Optional[int]:
        field = self.get_field(BikeProfileBikeSpdcadAntIdField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_spdcad_ant_id.setter
    def bike_spdcad_ant_id(self, value: int):
        field = self.get_field(BikeProfileBikeSpdcadAntIdField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_power_ant_id(self) -> Optional[int]:
        field = self.get_field(BikeProfileBikePowerAntIdField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_power_ant_id.setter
    def bike_power_ant_id(self, value: int):
        field = self.get_field(BikeProfileBikePowerAntIdField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def custom_wheelsize(self) -> Optional[float]:
        field = self.get_field(BikeProfileCustomWheelsizeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @custom_wheelsize.setter
    def custom_wheelsize(self, value: float):
        field = self.get_field(BikeProfileCustomWheelsizeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def auto_wheelsize(self) -> Optional[float]:
        field = self.get_field(BikeProfileAutoWheelsizeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @auto_wheelsize.setter
    def auto_wheelsize(self, value: float):
        field = self.get_field(BikeProfileAutoWheelsizeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_weight(self) -> Optional[float]:
        field = self.get_field(BikeProfileBikeWeightField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_weight.setter
    def bike_weight(self, value: float):
        field = self.get_field(BikeProfileBikeWeightField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def power_cal_factor(self) -> Optional[float]:
        field = self.get_field(BikeProfilePowerCalFactorField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @power_cal_factor.setter
    def power_cal_factor(self, value: float):
        field = self.get_field(BikeProfilePowerCalFactorField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def auto_wheel_cal(self) -> Optional[bool]:
        field = self.get_field(BikeProfileAutoWheelCalField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @auto_wheel_cal.setter
    def auto_wheel_cal(self, value: bool):
        field = self.get_field(BikeProfileAutoWheelCalField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def auto_power_zero(self) -> Optional[bool]:
        field = self.get_field(BikeProfileAutoPowerZeroField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @auto_power_zero.setter
    def auto_power_zero(self, value: bool):
        field = self.get_field(BikeProfileAutoPowerZeroField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def id(self) -> Optional[int]:
        field = self.get_field(BikeProfileIdField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @id.setter
    def id(self, value: int):
        field = self.get_field(BikeProfileIdField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def spd_enabled(self) -> Optional[bool]:
        field = self.get_field(BikeProfileSpdEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @spd_enabled.setter
    def spd_enabled(self, value: bool):
        field = self.get_field(BikeProfileSpdEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def cad_enabled(self) -> Optional[bool]:
        field = self.get_field(BikeProfileCadEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @cad_enabled.setter
    def cad_enabled(self, value: bool):
        field = self.get_field(BikeProfileCadEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def spdcad_enabled(self) -> Optional[bool]:
        field = self.get_field(BikeProfileSpdcadEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @spdcad_enabled.setter
    def spdcad_enabled(self, value: bool):
        field = self.get_field(BikeProfileSpdcadEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def power_enabled(self) -> Optional[bool]:
        field = self.get_field(BikeProfilePowerEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @power_enabled.setter
    def power_enabled(self, value: bool):
        field = self.get_field(BikeProfilePowerEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def crank_length(self) -> Optional[float]:
        field = self.get_field(BikeProfileCrankLengthField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @crank_length.setter
    def crank_length(self, value: float):
        field = self.get_field(BikeProfileCrankLengthField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def enabled(self) -> Optional[bool]:
        field = self.get_field(BikeProfileEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @enabled.setter
    def enabled(self, value: bool):
        field = self.get_field(BikeProfileEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_spd_ant_id_trans_type(self) -> Optional[int]:
        field = self.get_field(BikeProfileBikeSpdAntIdTransTypeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_spd_ant_id_trans_type.setter
    def bike_spd_ant_id_trans_type(self, value: int):
        field = self.get_field(BikeProfileBikeSpdAntIdTransTypeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_cad_ant_id_trans_type(self) -> Optional[int]:
        field = self.get_field(BikeProfileBikeCadAntIdTransTypeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_cad_ant_id_trans_type.setter
    def bike_cad_ant_id_trans_type(self, value: int):
        field = self.get_field(BikeProfileBikeCadAntIdTransTypeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_spdcad_ant_id_trans_type(self) -> Optional[int]:
        field = self.get_field(BikeProfileBikeSpdcadAntIdTransTypeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_spdcad_ant_id_trans_type.setter
    def bike_spdcad_ant_id_trans_type(self, value: int):
        field = self.get_field(BikeProfileBikeSpdcadAntIdTransTypeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bike_power_ant_id_trans_type(self) -> Optional[int]:
        field = self.get_field(BikeProfileBikePowerAntIdTransTypeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bike_power_ant_id_trans_type.setter
    def bike_power_ant_id_trans_type(self, value: int):
        field = self.get_field(BikeProfileBikePowerAntIdTransTypeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def odometer_rollover(self) -> Optional[int]:
        field = self.get_field(BikeProfileOdometerRolloverField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @odometer_rollover.setter
    def odometer_rollover(self, value: int):
        field = self.get_field(BikeProfileOdometerRolloverField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def front_gear_num(self) -> Optional[int]:
        field = self.get_field(BikeProfileFrontGearNumField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @front_gear_num.setter
    def front_gear_num(self, value: int):
        field = self.get_field(BikeProfileFrontGearNumField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def front_gear(self) -> Optional[list[int]]:
        field = self.get_field(BikeProfileFrontGearField.ID)
        if field and field.is_valid():
            return field.get_values()
        else:
            return None

    @front_gear.setter
    def front_gear(self, value: list[int]):
        field = self.get_field(BikeProfileFrontGearField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                field.set_values(value)

    @property
    def rear_gear_num(self) -> Optional[int]:
        field = self.get_field(BikeProfileRearGearNumField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @rear_gear_num.setter
    def rear_gear_num(self, value: int):
        field = self.get_field(BikeProfileRearGearNumField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def rear_gear(self) -> Optional[list[int]]:
        field = self.get_field(BikeProfileRearGearField.ID)
        if field and field.is_valid():
            return field.get_values()
        else:
            return None

    @rear_gear.setter
    def rear_gear(self, value: list[int]):
        field = self.get_field(BikeProfileRearGearField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                field.set_values(value)

    @property
    def shimano_di2_enabled(self) -> Optional[bool]:
        field = self.get_field(BikeProfileShimanoDi2EnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @shimano_di2_enabled.setter
    def shimano_di2_enabled(self, value: bool):
        field = self.get_field(BikeProfileShimanoDi2EnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)


class MessageIndexField(Field):
    ID = 254

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='message_index',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileNameField(Field):
    ID = 0

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='name',
            field_id=self.ID,
            base_type=BaseType.STRING,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileSportField(Field):
    ID = 1

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='sport',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileSubSportField(Field):
    ID = 2

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='sub_sport',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileOdometerField(Field):
    ID = 3

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='odometer',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=100,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileBikeSpdAntIdField(Field):
    ID = 4

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bike_spd_ant_id',
            field_id=self.ID,
            base_type=BaseType.UINT16Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileBikeCadAntIdField(Field):
    ID = 5

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bike_cad_ant_id',
            field_id=self.ID,
            base_type=BaseType.UINT16Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileBikeSpdcadAntIdField(Field):
    ID = 6

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bike_spdcad_ant_id',
            field_id=self.ID,
            base_type=BaseType.UINT16Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileBikePowerAntIdField(Field):
    ID = 7

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bike_power_ant_id',
            field_id=self.ID,
            base_type=BaseType.UINT16Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileCustomWheelsizeField(Field):
    ID = 8

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='custom_wheelsize',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1000,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileAutoWheelsizeField(Field):
    ID = 9

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='auto_wheelsize',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1000,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileBikeWeightField(Field):
    ID = 10

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bike_weight',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=10,
            size=size,
            units='kg',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfilePowerCalFactorField(Field):
    ID = 11

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='power_cal_factor',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=10,
            size=size,
            units='%',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileAutoWheelCalField(Field):
    ID = 12

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='auto_wheel_cal',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileAutoPowerZeroField(Field):
    ID = 13

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='auto_power_zero',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileIdField(Field):
    ID = 14

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='id',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileSpdEnabledField(Field):
    ID = 15

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='spd_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileCadEnabledField(Field):
    ID = 16

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='cad_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileSpdcadEnabledField(Field):
    ID = 17

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='spdcad_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfilePowerEnabledField(Field):
    ID = 18

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='power_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileCrankLengthField(Field):
    ID = 19

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='crank_length',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=-110,
            scale=2,
            size=size,
            units='mm',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileEnabledField(Field):
    ID = 20

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileBikeSpdAntIdTransTypeField(Field):
    ID = 21

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bike_spd_ant_id_trans_type',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileBikeCadAntIdTransTypeField(Field):
    ID = 22

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bike_cad_ant_id_trans_type',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileBikeSpdcadAntIdTransTypeField(Field):
    ID = 23

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bike_spdcad_ant_id_trans_type',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileBikePowerAntIdTransTypeField(Field):
    ID = 24

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bike_power_ant_id_trans_type',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileOdometerRolloverField(Field):
    ID = 37

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='odometer_rollover',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileFrontGearNumField(Field):
    ID = 38

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='front_gear_num',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileFrontGearField(Field):
    ID = 39

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='front_gear',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileRearGearNumField(Field):
    ID = 40

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='rear_gear_num',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileRearGearField(Field):
    ID = 41

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='rear_gear',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class BikeProfileShimanoDi2EnabledField(Field):
    ID = 44

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='shimano_di2_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )



import csv
import struct
from typing import List as list

from fit_tool.base_type import BaseType
from fit_tool.developer_field import DeveloperField
from fit_tool.fit_file_header import FitFileHeader
from fit_tool.profile.messages.field_description_message import FieldDescriptionMessage
from fit_tool.record import Record
from fit_tool.utils.crc import crc16
from fit_tool.utils.logging import logger


class FitFile:
    def __init__(self, header: FitFileHeader, records: list[Record], crc: int = None):
        self.header = header
        self.records = records
        self.crc = crc  # crc16 of header and records

    @classmethod
    def from_file(cls, path: str):
        with open(path, 'rb') as file_object:
            bytes_buffer = file_object.read()
            fit_file = FitFile.from_bytes(bytes_buffer)
            return fit_file

    @classmethod
    def from_bytes(cls, bytes_buffer: bytes, check_crc: bool = True):
        crc = 0
        offset = 0

        header_size = bytes_buffer[0]

        header_bytes = bytes_buffer[:header_size]
        header = FitFileHeader.from_bytes(header_bytes)
        crc = crc16(header_bytes, crc=crc)
        offset += header_size

        records = []
        definition_messages = {}
        developer_fields_by_data_index = {}

        record_index = 0
        record_bytes_remaining_count = header.records_size
        while record_bytes_remaining_count > 0:
            record = Record.from_bytes(definition_messages=definition_messages, bytes_buffer=bytes_buffer,
                                       offset=offset, developer_fields_by_data_index=developer_fields_by_data_index)

            if record.is_definition:
                definition_messages[record.local_id] = record.message
            elif isinstance(record.message, FieldDescriptionMessage):
                message = record.message
                developer_field = DeveloperField(developer_data_index=message.developer_data_index,
                                                 field_id=message.field_definition_number,
                                                 base_type=BaseType(message.fit_base_type_id),
                                                 name=message.field_name,
                                                 scale=message.scale,
                                                 offset=message.offset,
                                                 units=message.units)
                if developer_field.developer_data_index not in developer_fields_by_data_index:
                    developer_fields_by_data_index[developer_field.developer_data_index] = {}

                developer_fields_by_data_index[developer_field.developer_data_index][
                    developer_field.field_id] = developer_field

            records.append(record)
            definition_message = definition_messages[record.local_id]
            record_size = record.size
            defined_size = record.defined_size(definition_message)
            crc = crc16(bytes_buffer[offset:offset + defined_size], crc=crc)

            if record_size != defined_size:
                logger.warning(
                    f'Record {record_index}, {record.message}: size ({record_size}) != defined size ({defined_size}). Some fields were not read correctly.')

            actual_bytes = bytes_buffer[offset:offset + defined_size]
            record_bytes = record.to_bytes()

            if actual_bytes != record_bytes:
                logger.warning(f'- {record_index} -\n\tactual: {actual_bytes}\n\trecord: {record_bytes}')

            record_bytes_remaining_count -= defined_size
            offset += defined_size
            record_index += 1

        file_crc, = struct.unpack(f'<H', bytes_buffer[offset:offset + 2])

        if crc != file_crc:
            message = f'Calculated crc ({hex(crc)}) does match crc in file ({hex(file_crc)}).'

            if check_crc:
                raise Exception(message)
            else:
                logger.warning(message)

        return FitFile(header, records, crc)

    def to_bytes(self, check_crc: bool = True):
        calculated_crc = 0
        bytes_buffer = bytearray()
        buffer = self.header.to_bytes()
        calculated_crc = crc16(buffer, crc=calculated_crc)
        bytes_buffer.extend(buffer)

        for record in self.records:
            buffer = record.to_bytes()
            calculated_crc = crc16(buffer, crc=calculated_crc)
            bytes_buffer.extend(buffer)

        if self.crc is None:
            self.crc = calculated_crc

        if self.crc != calculated_crc:
            message = f'Calculated crc ({calculated_crc}) != defined crc ({self.crc})'

            if check_crc:
                raise Exception(message)
            else:
                logger.warn(message)

        buffer = struct.pack('<H', self.crc)
        bytes_buffer.extend(buffer)

        return bytes(bytes_buffer)

    def to_rows(self) -> list[list]:
        result = []

        max_columns = 0
        for record in self.records:
            row = record.to_row()
            max_columns = max(max_columns, len(row))
            result.append(row)

        header_row = ['Type', 'Local ID', 'Message']
        max_fields = (max_columns - 3) // 3

        for i in range(max_fields):
            header_row.extend([f'Field {i}', f'Value {i}', f'Units {i}'])

        result.insert(0, header_row)

        return result

    def to_csv(self, path: str):

        rows = self.to_rows()

        with open(path, 'w', newline='') as csv_file:
            csv_writer = csv.writer(csv_file, delimiter=',',
                                    quotechar='"', quoting=csv.QUOTE_MINIMAL)
            for row in rows:
                csv_writer.writerow(row)

    def to_file(self, path: str):
        with open(path, 'wb') as file_object:
            file_object.write(self.to_bytes())



# milliseconds from Unix epoch (start of 1970) to Sun Dec 31 00:00:00 1989
MILLISECONDS_EPOCH_1989_DELTA = 631065600000


def to_seconds_since_1989_epoch(milliseconds_since_epoch: int) -> int:
    seconds_since_1989_epoch = (milliseconds_since_epoch - MILLISECONDS_EPOCH_1989_DELTA) // 1000
    return seconds_since_1989_epoch


def to_milliseconds_since_epoch(seconds_since_1989_epoch: int) -> int:
    milliseconds_since_epoch = seconds_since_1989_epoch * 1000 + MILLISECONDS_EPOCH_1989_DELTA
    return milliseconds_since_epoch


def to_semicircles(degrees: int) -> int:
    return round(2147483648 * degrees / 180.0)


def to_degrees(semicircles: int) -> float:
    return (semicircles * 180.0) / 2147483648



# Autogenerated. Do not modify.
#
# Profile: 21.60
from typing import List as list
from typing import Optional

from fit_tool.base_type import BaseType
from fit_tool.data_message import DataMessage
from fit_tool.definition_message import DefinitionMessage
from fit_tool.developer_field import DeveloperField
from fit_tool.endian import Endian
from fit_tool.field import Field
from fit_tool.profile.profile_type import *


class ExdDataConceptConfigurationMessage(DataMessage):
    ID = 202
    NAME = 'exd_data_concept_configuration'

    @staticmethod
    def __get_field_size(definition_message: DefinitionMessage, field_id: int) -> int:
        size = 0
        if definition_message:
            field_definition = definition_message.get_field_definition(field_id)
            if field_definition:
                size = field_definition.size

        return size

    def __init__(self, definition_message=None, developer_fields=None, local_id: int = 0,
                 endian: Endian = Endian.LITTLE):
        super().__init__(name=ExdDataConceptConfigurationMessage.NAME,
                         global_id=ExdDataConceptConfigurationMessage.ID,
                         local_id=definition_message.local_id if definition_message else local_id,
                         endian=definition_message.endian if definition_message else endian,
                         definition_message=definition_message,
                         developer_fields=developer_fields,
                         fields=[
                             ExdDataConceptConfigurationScreenIndexField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationScreenIndexField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationConceptField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationConceptField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationFieldIdField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationFieldIdField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationConceptIndexField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationConceptIndexField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationDataPageField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationDataPageField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationConceptKeyField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationConceptKeyField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationScalingField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationScalingField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationDataUnitsField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationDataUnitsField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationQualifierField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationQualifierField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationDescriptorField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationDescriptorField.ID),
                                 growable=definition_message is None),
                             ExdDataConceptConfigurationIsSignedField(
                                 size=self.__get_field_size(definition_message,
                                                            ExdDataConceptConfigurationIsSignedField.ID),
                                 growable=definition_message is None)
                         ])

        self.growable = self.definition_message is None

    @classmethod
    def from_bytes(cls, definition_message: DefinitionMessage, developer_fields: list[DeveloperField],
                   bytes_buffer: bytes, offset: int = 0):
        message = cls(definition_message=definition_message, developer_fields=developer_fields)
        message.read_from_bytes(bytes_buffer, offset)
        return message

    @property
    def screen_index(self) -> Optional[int]:
        field = self.get_field(ExdDataConceptConfigurationScreenIndexField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @screen_index.setter
    def screen_index(self, value: int):
        field = self.get_field(ExdDataConceptConfigurationScreenIndexField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def concept_field(self) -> Optional[int]:
        field = self.get_field(ExdDataConceptConfigurationConceptField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @concept_field.setter
    def concept_field(self, value: int):
        field = self.get_field(ExdDataConceptConfigurationConceptField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def field_id(self) -> Optional[int]:
        field = self.get_field(ExdDataConceptConfigurationFieldIdField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @field_id.setter
    def field_id(self, value: int):
        field = self.get_field(ExdDataConceptConfigurationFieldIdField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def concept_index(self) -> Optional[int]:
        field = self.get_field(ExdDataConceptConfigurationConceptIndexField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @concept_index.setter
    def concept_index(self, value: int):
        field = self.get_field(ExdDataConceptConfigurationConceptIndexField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def data_page(self) -> Optional[int]:
        field = self.get_field(ExdDataConceptConfigurationDataPageField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @data_page.setter
    def data_page(self, value: int):
        field = self.get_field(ExdDataConceptConfigurationDataPageField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def concept_key(self) -> Optional[int]:
        field = self.get_field(ExdDataConceptConfigurationConceptKeyField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @concept_key.setter
    def concept_key(self, value: int):
        field = self.get_field(ExdDataConceptConfigurationConceptKeyField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def scaling(self) -> Optional[int]:
        field = self.get_field(ExdDataConceptConfigurationScalingField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @scaling.setter
    def scaling(self, value: int):
        field = self.get_field(ExdDataConceptConfigurationScalingField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def data_units(self) -> Optional[ExdDataUnits]:
        field = self.get_field(ExdDataConceptConfigurationDataUnitsField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @data_units.setter
    def data_units(self, value: ExdDataUnits):
        field = self.get_field(ExdDataConceptConfigurationDataUnitsField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def qualifier(self) -> Optional[ExdQualifiers]:
        field = self.get_field(ExdDataConceptConfigurationQualifierField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @qualifier.setter
    def qualifier(self, value: ExdQualifiers):
        field = self.get_field(ExdDataConceptConfigurationQualifierField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def descriptor(self) -> Optional[ExdDescriptors]:
        field = self.get_field(ExdDataConceptConfigurationDescriptorField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @descriptor.setter
    def descriptor(self, value: ExdDescriptors):
        field = self.get_field(ExdDataConceptConfigurationDescriptorField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def is_signed(self) -> Optional[bool]:
        field = self.get_field(ExdDataConceptConfigurationIsSignedField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @is_signed.setter
    def is_signed(self, value: bool):
        field = self.get_field(ExdDataConceptConfigurationIsSignedField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)


class ExdDataConceptConfigurationScreenIndexField(Field):
    ID = 0

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='screen_index',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationConceptField(Field):
    ID = 1

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='concept_field',
            field_id=self.ID,
            base_type=BaseType.BYTE,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationFieldIdField(Field):
    ID = 2

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='field_id',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationConceptIndexField(Field):
    ID = 3

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='concept_index',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationDataPageField(Field):
    ID = 4

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='data_page',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationConceptKeyField(Field):
    ID = 5

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='concept_key',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationScalingField(Field):
    ID = 6

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='scaling',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationDataUnitsField(Field):
    ID = 8

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='data_units',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationQualifierField(Field):
    ID = 9

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='qualifier',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationDescriptorField(Field):
    ID = 10

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='descriptor',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ExdDataConceptConfigurationIsSignedField(Field):
    ID = 11

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='is_signed',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )



# Autogenerated. Do not modify.
#
# Profile: 21.60
from typing import List as list
from typing import Optional

from fit_tool.base_type import BaseType
from fit_tool.data_message import DataMessage
from fit_tool.definition_message import DefinitionMessage
from fit_tool.developer_field import DeveloperField
from fit_tool.endian import Endian
from fit_tool.field import Field
from fit_tool.profile.profile_type import *


class SegmentIdMessage(DataMessage):
    ID = 148
    NAME = 'segment_id'

    @staticmethod
    def __get_field_size(definition_message: DefinitionMessage, field_id: int) -> int:
        size = 0
        if definition_message:
            field_definition = definition_message.get_field_definition(field_id)
            if field_definition:
                size = field_definition.size

        return size

    def __init__(self, definition_message=None, developer_fields=None, local_id: int = 0,
                 endian: Endian = Endian.LITTLE):
        super().__init__(name=SegmentIdMessage.NAME,
                         global_id=SegmentIdMessage.ID,
                         local_id=definition_message.local_id if definition_message else local_id,
                         endian=definition_message.endian if definition_message else endian,
                         definition_message=definition_message,
                         developer_fields=developer_fields,
                         fields=[
                             SegmentIdNameField(
                                 size=self.__get_field_size(definition_message, SegmentIdNameField.ID),
                                 growable=definition_message is None),
                             SegmentIdUuidField(
                                 size=self.__get_field_size(definition_message, SegmentIdUuidField.ID),
                                 growable=definition_message is None),
                             SegmentIdSportField(
                                 size=self.__get_field_size(definition_message, SegmentIdSportField.ID),
                                 growable=definition_message is None),
                             SegmentIdEnabledField(
                                 size=self.__get_field_size(definition_message, SegmentIdEnabledField.ID),
                                 growable=definition_message is None),
                             SegmentIdUserProfilePrimaryKeyField(
                                 size=self.__get_field_size(definition_message, SegmentIdUserProfilePrimaryKeyField.ID),
                                 growable=definition_message is None),
                             SegmentIdDeviceIdField(
                                 size=self.__get_field_size(definition_message, SegmentIdDeviceIdField.ID),
                                 growable=definition_message is None),
                             SegmentIdDefaultRaceLeaderField(
                                 size=self.__get_field_size(definition_message, SegmentIdDefaultRaceLeaderField.ID),
                                 growable=definition_message is None),
                             SegmentIdDeleteStatusField(
                                 size=self.__get_field_size(definition_message, SegmentIdDeleteStatusField.ID),
                                 growable=definition_message is None),
                             SegmentIdSelectionTypeField(
                                 size=self.__get_field_size(definition_message, SegmentIdSelectionTypeField.ID),
                                 growable=definition_message is None)
                         ])

        self.growable = self.definition_message is None

    @classmethod
    def from_bytes(cls, definition_message: DefinitionMessage, developer_fields: list[DeveloperField],
                   bytes_buffer: bytes, offset: int = 0):
        message = cls(definition_message=definition_message, developer_fields=developer_fields)
        message.read_from_bytes(bytes_buffer, offset)
        return message

    @property
    def segment_id_name(self) -> Optional[str]:
        field = self.get_field(SegmentIdNameField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @segment_id_name.setter
    def segment_id_name(self, value: str):
        field = self.get_field(SegmentIdNameField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def uuid(self) -> Optional[str]:
        field = self.get_field(SegmentIdUuidField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @uuid.setter
    def uuid(self, value: str):
        field = self.get_field(SegmentIdUuidField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def sport(self) -> Optional[Sport]:
        field = self.get_field(SegmentIdSportField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @sport.setter
    def sport(self, value: Sport):
        field = self.get_field(SegmentIdSportField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def enabled(self) -> Optional[bool]:
        field = self.get_field(SegmentIdEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @enabled.setter
    def enabled(self, value: bool):
        field = self.get_field(SegmentIdEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def user_profile_primary_key(self) -> Optional[int]:
        field = self.get_field(SegmentIdUserProfilePrimaryKeyField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @user_profile_primary_key.setter
    def user_profile_primary_key(self, value: int):
        field = self.get_field(SegmentIdUserProfilePrimaryKeyField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def device_id(self) -> Optional[int]:
        field = self.get_field(SegmentIdDeviceIdField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @device_id.setter
    def device_id(self, value: int):
        field = self.get_field(SegmentIdDeviceIdField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def default_race_leader(self) -> Optional[int]:
        field = self.get_field(SegmentIdDefaultRaceLeaderField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @default_race_leader.setter
    def default_race_leader(self, value: int):
        field = self.get_field(SegmentIdDefaultRaceLeaderField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def delete_status(self) -> Optional[SegmentDeleteStatus]:
        field = self.get_field(SegmentIdDeleteStatusField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @delete_status.setter
    def delete_status(self, value: SegmentDeleteStatus):
        field = self.get_field(SegmentIdDeleteStatusField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def selection_type(self) -> Optional[SegmentSelectionType]:
        field = self.get_field(SegmentIdSelectionTypeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @selection_type.setter
    def selection_type(self, value: SegmentSelectionType):
        field = self.get_field(SegmentIdSelectionTypeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)


class SegmentIdNameField(Field):
    ID = 0

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='name',
            field_id=self.ID,
            base_type=BaseType.STRING,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class SegmentIdUuidField(Field):
    ID = 1

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='uuid',
            field_id=self.ID,
            base_type=BaseType.STRING,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class SegmentIdSportField(Field):
    ID = 2

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='sport',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class SegmentIdEnabledField(Field):
    ID = 3

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class SegmentIdUserProfilePrimaryKeyField(Field):
    ID = 4

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='user_profile_primary_key',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class SegmentIdDeviceIdField(Field):
    ID = 5

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='device_id',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class SegmentIdDefaultRaceLeaderField(Field):
    ID = 6

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='default_race_leader',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class SegmentIdDeleteStatusField(Field):
    ID = 7

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='delete_status',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class SegmentIdSelectionTypeField(Field):
    ID = 8

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='selection_type',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )



# Autogenerated. Do not modify.
#
# Profile: 21.60
from typing import List as list
from typing import Optional

from fit_tool.base_type import BaseType
from fit_tool.data_message import DataMessage
from fit_tool.definition_message import DefinitionMessage
from fit_tool.developer_field import DeveloperField
from fit_tool.endian import Endian
from fit_tool.field import Field
from fit_tool.profile.profile_type import *


class GoalMessage(DataMessage):
    ID = 15
    NAME = 'goal'

    @staticmethod
    def __get_field_size(definition_message: DefinitionMessage, field_id: int) -> int:
        size = 0
        if definition_message:
            field_definition = definition_message.get_field_definition(field_id)
            if field_definition:
                size = field_definition.size

        return size

    def __init__(self, definition_message=None, developer_fields=None, local_id: int = 0,
                 endian: Endian = Endian.LITTLE):
        super().__init__(name=GoalMessage.NAME,
                         global_id=GoalMessage.ID,
                         local_id=definition_message.local_id if definition_message else local_id,
                         endian=definition_message.endian if definition_message else endian,
                         definition_message=definition_message,
                         developer_fields=developer_fields,
                         fields=[
                             MessageIndexField(
                                 size=self.__get_field_size(definition_message, MessageIndexField.ID),
                                 growable=definition_message is None),
                             GoalSportField(
                                 size=self.__get_field_size(definition_message, GoalSportField.ID),
                                 growable=definition_message is None),
                             GoalSubSportField(
                                 size=self.__get_field_size(definition_message, GoalSubSportField.ID),
                                 growable=definition_message is None),
                             GoalStartDateField(
                                 size=self.__get_field_size(definition_message, GoalStartDateField.ID),
                                 growable=definition_message is None),
                             GoalEndDateField(
                                 size=self.__get_field_size(definition_message, GoalEndDateField.ID),
                                 growable=definition_message is None),
                             GoalTypeField(
                                 size=self.__get_field_size(definition_message, GoalTypeField.ID),
                                 growable=definition_message is None),
                             GoalValueField(
                                 size=self.__get_field_size(definition_message, GoalValueField.ID),
                                 growable=definition_message is None),
                             GoalRepeatField(
                                 size=self.__get_field_size(definition_message, GoalRepeatField.ID),
                                 growable=definition_message is None),
                             GoalTargetValueField(
                                 size=self.__get_field_size(definition_message, GoalTargetValueField.ID),
                                 growable=definition_message is None),
                             GoalRecurrenceField(
                                 size=self.__get_field_size(definition_message, GoalRecurrenceField.ID),
                                 growable=definition_message is None),
                             GoalRecurrenceValueField(
                                 size=self.__get_field_size(definition_message, GoalRecurrenceValueField.ID),
                                 growable=definition_message is None),
                             GoalEnabledField(
                                 size=self.__get_field_size(definition_message, GoalEnabledField.ID),
                                 growable=definition_message is None),
                             GoalSourceField(
                                 size=self.__get_field_size(definition_message, GoalSourceField.ID),
                                 growable=definition_message is None)
                         ])

        self.growable = self.definition_message is None

    @classmethod
    def from_bytes(cls, definition_message: DefinitionMessage, developer_fields: list[DeveloperField],
                   bytes_buffer: bytes, offset: int = 0):
        message = cls(definition_message=definition_message, developer_fields=developer_fields)
        message.read_from_bytes(bytes_buffer, offset)
        return message

    @property
    def message_index(self) -> Optional[int]:
        field = self.get_field(MessageIndexField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @message_index.setter
    def message_index(self, value: int):
        field = self.get_field(MessageIndexField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def sport(self) -> Optional[Sport]:
        field = self.get_field(GoalSportField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @sport.setter
    def sport(self, value: Sport):
        field = self.get_field(GoalSportField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def sub_sport(self) -> Optional[SubSport]:
        field = self.get_field(GoalSubSportField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @sub_sport.setter
    def sub_sport(self, value: SubSport):
        field = self.get_field(GoalSubSportField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @property
    def start_date(self) -> Optional[int]:
        field = self.get_field(GoalStartDateField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @start_date.setter
    def start_date(self, value: int):
        field = self.get_field(GoalStartDateField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @property
    def end_date(self) -> Optional[int]:
        field = self.get_field(GoalEndDateField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @end_date.setter
    def end_date(self, value: int):
        field = self.get_field(GoalEndDateField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def type(self) -> Optional[Goal]:
        field = self.get_field(GoalTypeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @type.setter
    def type(self, value: Goal):
        field = self.get_field(GoalTypeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def value(self) -> Optional[int]:
        field = self.get_field(GoalValueField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @value.setter
    def value(self, value: int):
        field = self.get_field(GoalValueField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def repeat(self) -> Optional[bool]:
        field = self.get_field(GoalRepeatField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @repeat.setter
    def repeat(self, value: bool):
        field = self.get_field(GoalRepeatField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def target_value(self) -> Optional[int]:
        field = self.get_field(GoalTargetValueField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @target_value.setter
    def target_value(self, value: int):
        field = self.get_field(GoalTargetValueField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def recurrence(self) -> Optional[GoalRecurrence]:
        field = self.get_field(GoalRecurrenceField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @recurrence.setter
    def recurrence(self, value: GoalRecurrence):
        field = self.get_field(GoalRecurrenceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def recurrence_value(self) -> Optional[int]:
        field = self.get_field(GoalRecurrenceValueField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @recurrence_value.setter
    def recurrence_value(self, value: int):
        field = self.get_field(GoalRecurrenceValueField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def enabled(self) -> Optional[bool]:
        field = self.get_field(GoalEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @enabled.setter
    def enabled(self, value: bool):
        field = self.get_field(GoalEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def source(self) -> Optional[GoalSource]:
        field = self.get_field(GoalSourceField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @source.setter
    def source(self, value: GoalSource):
        field = self.get_field(GoalSourceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)


class MessageIndexField(Field):
    ID = 254

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='message_index',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalSportField(Field):
    ID = 0

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='sport',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalSubSportField(Field):
    ID = 1

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='sub_sport',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalStartDateField(Field):
    ID = 2

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='start_date',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=-631065600000,
            scale=0.001,
            size=size,
            units='ms',
            type_name='date_time',
            growable=growable,
            sub_fields=[
            ]
        )


class GoalEndDateField(Field):
    ID = 3

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='end_date',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=-631065600000,
            scale=0.001,
            size=size,
            units='ms',
            type_name='date_time',
            growable=growable,
            sub_fields=[
            ]
        )


class GoalTypeField(Field):
    ID = 4

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='type',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalValueField(Field):
    ID = 5

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='value',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalRepeatField(Field):
    ID = 6

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='repeat',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalTargetValueField(Field):
    ID = 7

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='target_value',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalRecurrenceField(Field):
    ID = 8

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='recurrence',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalRecurrenceValueField(Field):
    ID = 9

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='recurrence_value',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalEnabledField(Field):
    ID = 10

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class GoalSourceField(Field):
    ID = 11

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='source',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )



import struct

from fit_tool.base_type import BaseType
from fit_tool.developer_field import DeveloperField


class DeveloperFieldDefinition:
    PACK_FORMAT = 'BBB'

    def __init__(self, field_id: int, size: int, developer_data_index: int):
        self.field_id = field_id
        self.size = size
        self.developer_data_index = developer_data_index

    def to_bytes(self):
        return struct.pack(self.PACK_FORMAT, self.field_id, self.size,
                           self.developer_data_index)

    @classmethod
    def from_bytes(cls, bytes_buffer, offset=0):
        field_id, size, developer_data_index = struct.unpack_from(cls.PACK_FORMAT, bytes_buffer, offset)

        return cls(field_id, size, developer_data_index)

    @classmethod
    def field_definition_size(cls):
        return struct.calcsize(cls.PACK_FORMAT)

    @classmethod
    def from_field(cls, field: DeveloperField, min_string_size: int = 0):
        if field.base_type == BaseType.STRING:
            size = max(field.size, min_string_size)
        else:
            size = field.size

        return cls(field_id=field.field_id, size=size, developer_data_index=field.developer_data_index)
